#include "sparrowrenderer.h"
#include "guipipeline.h"
#include "texture.h"
#include "scene.h"
#include "mesh.h"
#include "shader.h"
#include <glm/ext.hpp>
#include "shadersource.h"
#include <resource.h>

RESOURCE_PACK(shaders)

GuiBuffer::GuiBuffer(int width, int height) : FrameBuffer()
{
    bindFBO();
    glGenRenderbuffers(1, &m_rbo);
    glBindRenderbuffer(GL_RENDERBUFFER, m_rbo);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT32, width, height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, m_rbo);
    Texture* tex = new Texture(GL_RGBA, GL_RGBA8UI, width, height, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE);
    addTexture(tex, GL_COLOR_ATTACHMENT0);
    initColorAttachments();
}

GuiBuffer::~GuiBuffer()
{
    glDeleteRenderbuffers(1, &m_rbo);
}

GuiPipeline::GuiPipeline() :
    m_width(512),
    m_height(512),
    m_guiBuffer(NULL)
{
    Resource::ResourceMap shaderMap;
    Resource::getResourcePack_shaders(shaderMap);
    m_guiShader = new ShaderSource();
    m_guiShader->setSource(shaderMap["shaders/gui.vert.glsl"], ShaderSource::VERTEX);
    m_guiShader->setSource(shaderMap["shaders/gui.frag.glsl"], ShaderSource::FRAGMENT);
    m_renderTarget = FrameBuffer::screen;
}

void GuiPipeline::renderGL(Scene *scene)
{
    if(m_renderTarget == NULL)
        return;

    // GEOMETRY PASS
    m_renderTarget->bindFBO();
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);
    glClearColor(0.0f, 0.0f, 0.0f, 1.f);
    glClearDepth(1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    // loop on geometry
    for(SceneIterator<GeometryNode*>* geometryIt = scene->getGeometry();
        geometryIt->isValid(); geometryIt->next())
    {
        
        GeometryNode* node = geometryIt->getItem();
        Shader *shader = m_meshShaders[node->mesh->getFlags()];
        if(shader == NULL)
            continue;
        shader->bind();
        
        shader->bindFloat(shader->getLocation("depth"), node->mesh->getDepth());
        shader->bindMat4(shader->getLocation("transformMatrix"), node->modelMatrix);
        shader->bindMat4(shader->getLocation("orthoMatrix"), m_orthoMatrix);

        // draw geometry
        node->mesh->draw(shader);
    }
}

void GuiPipeline::resizeGL(int w, int h)
{
    // updating dimensions
    m_width = w;
    m_height = h;
    
    // rebuilding FrameBuffers
    /*if(m_guiBuffer != NULL)
        delete m_guiBuffer;
    m_guiBuffer = new GuiBuffer(w, h);*/
    m_orthoMatrix = glm::ortho(0.f, float(m_width), float(m_height), 0.f);
}

void GuiPipeline::setSources(ShaderSource *guiSource)
{
    if(m_guiShader != NULL)
        delete m_guiShader;
    m_guiShader = guiSource;
}

void GuiPipeline::refreshScene(Scene *scene)
{
    for(auto it : m_meshShaders)
        delete it.second;
    m_meshShaders.clear();
    m_meshTypes.clear();
    scene->getMeshTypes(m_meshTypes);
    for(unsigned int type : m_meshTypes)
    {
        if(type & (1 << Mesh::MESH_2D))
            m_meshShaders[type] = m_guiShader->compile(type, 0);
        else
            m_meshShaders[type] = NULL;
    }
}
